package ctbrec.recorder.download.hls;

import static ctbrec.io.HttpConstants.*;
import static ctbrec.io.HttpConstants.ORIGIN;
import static ctbrec.recorder.download.StreamSource.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import javax.xml.bind.JAXBException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.iheartradio.m3u8.Encoding;
import com.iheartradio.m3u8.Format;
import com.iheartradio.m3u8.ParseException;
import com.iheartradio.m3u8.ParsingMode;
import com.iheartradio.m3u8.PlaylistException;
import com.iheartradio.m3u8.PlaylistParser;
import com.iheartradio.m3u8.data.EncryptionData;
import com.iheartradio.m3u8.data.MediaPlaylist;
import com.iheartradio.m3u8.data.Playlist;
import com.iheartradio.m3u8.data.TrackData;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.Recording.State;
import ctbrec.UnknownModel;
import ctbrec.io.BandwidthMeter;
import ctbrec.io.HttpClient;
import ctbrec.io.HttpException;
import ctbrec.recorder.PlaylistGenerator.InvalidPlaylistException;
import ctbrec.recorder.download.AbstractDownload;
import ctbrec.recorder.download.HttpHeaderFactory;
import ctbrec.recorder.download.StreamSource;
import ctbrec.sites.Site;
import okhttp3.Request;
import okhttp3.Request.Builder;
import okhttp3.Response;

public abstract class AbstractHlsDownload extends AbstractDownload {

    private static final Logger LOG = LoggerFactory.getLogger(AbstractHlsDownload.class);
    private static int threadCounter = 0;

    protected transient HttpClient client;
    protected volatile boolean running = false;
    protected Model model = new UnknownModel();
    protected transient LinkedBlockingQueue<Runnable> downloadQueue = new LinkedBlockingQueue<>(50);
    protected transient ExecutorService downloadThreadPool = new ThreadPoolExecutor(0, 5, 20, TimeUnit.SECONDS, downloadQueue, createThreadFactory());
    protected State state = State.UNKNOWN;
    private int playlistEmptyCount = 0;

    protected AbstractHlsDownload(HttpClient client) {
        this.client = client;
    }

    private ThreadFactory createThreadFactory() {
        return r -> {
            Thread t = new Thread(r);
            t.setDaemon(true);
            t.setName("SegmentDownloadThread-" + threadCounter++);
            return t;
        };
    }

    protected SegmentPlaylist getNextSegments(String segmentsURL) throws IOException, ParseException, PlaylistException {
        URL segmentsUrl = new URL(segmentsURL);
        Builder builder = new Request.Builder()
                .url(segmentsUrl);
        addHeaders(builder, Optional.ofNullable(model).map(Model::getHttpHeaderFactory).map(HttpHeaderFactory::createSegmentPlaylistHeaders).orElse(new HashMap<>()));
        Request request = builder.build();

        try (Response response = client.execute(request)) {
            if (response.isSuccessful()) {
                String body = response.body().string();
                if (!body.contains("#EXTINF")) {
                    // no segments, empty playlist
                    return new SegmentPlaylist(segmentsURL);
                }

                byte[] bytes = body.getBytes(StandardCharsets.UTF_8);
                BandwidthMeter.add(bytes.length);
                InputStream inputStream = new ByteArrayInputStream(bytes);
                PlaylistParser parser = new PlaylistParser(inputStream, Format.EXT_M3U, Encoding.UTF_8, ParsingMode.LENIENT);
                Playlist playlist = parser.parse();
                if (playlist.hasMediaPlaylist()) {
                    MediaPlaylist mediaPlaylist = playlist.getMediaPlaylist();
                    SegmentPlaylist lsp = new SegmentPlaylist(segmentsURL);
                    lsp.seq = mediaPlaylist.getMediaSequenceNumber();
                    lsp.targetDuration = mediaPlaylist.getTargetDuration();
                    List<TrackData> tracks = mediaPlaylist.getTracks();
                    for (TrackData trackData : tracks) {
                        String uri = trackData.getUri();
                        if (!uri.startsWith("http")) {
                            URL context = new URL(segmentsURL);
                            uri = new URL(context, uri).toExternalForm();
                        }
                        lsp.totalDuration += trackData.getTrackInfo().duration;
                        lsp.lastSegDuration = trackData.getTrackInfo().duration;
                        lsp.segments.add(uri);
                        if (trackData.hasEncryptionData()) {
                            lsp.encrypted = true;
                            EncryptionData data = trackData.getEncryptionData();
                            lsp.encryptionKeyUrl = data.getUri();
                            lsp.encryptionMethod = data.getMethod().getValue();
                        }
                    }
                    return lsp;
                }
                throw new InvalidPlaylistException("Playlist has no media playlist");
            } else {
                throw new HttpException(response.code(), response.message());
            }
        }
    }


    protected void addHeaders(Builder builder, Map<String, String> headers) {
        headers.putIfAbsent(ACCEPT, "*/*");
        headers.putIfAbsent(ACCEPT_LANGUAGE, Locale.ENGLISH.getLanguage());
        headers.putIfAbsent(USER_AGENT, Config.getInstance().getSettings().httpUserAgent);
        headers.putIfAbsent(CONNECTION, KEEP_ALIVE);
        headers.computeIfAbsent(ORIGIN, k -> Optional.ofNullable(model).map(Model::getSite).map(Site::getBaseUrl).orElse(null));
        headers.computeIfAbsent(REFERER, k -> Optional.ofNullable(model).map(Model::getSite).map(Site::getBaseUrl).orElse(null));

        for (Entry<String, String> header : headers.entrySet()) {
            builder.header(header.getKey(), header.getValue());
        }
    }

    protected String getSegmentPlaylistUrl(Model model) throws IOException, ExecutionException, ParseException, PlaylistException, JAXBException {
        LOG.debug("{} stream idx: {}", model.getName(), model.getStreamUrlIndex());
        List<StreamSource> streamSources = model.getStreamSources();
        Collections.sort(streamSources);
        for (StreamSource streamSource : streamSources) {
            LOG.debug("{} src {}", model.getName(), streamSource);
        }
        String url = null;
        if (model.getStreamUrlIndex() >= 0 && model.getStreamUrlIndex() < streamSources.size()) {
            // TODO don't use the index, but the bandwidth. if the bandwidth does not match, take the closest one
            LOG.debug("{} selected {}", model.getName(), streamSources.get(model.getStreamUrlIndex()));
            url = streamSources.get(model.getStreamUrlIndex()).getMediaPlaylistUrl();
        } else {
            // filter out stream resolutions, which are out of range of the configured min and max
            int minRes = Config.getInstance().getSettings().minimumResolution;
            int maxRes = Config.getInstance().getSettings().maximumResolution;
            List<StreamSource> filteredStreamSources = streamSources.stream()
                    .filter(src -> src.height == 0 || src.height == UNKNOWN || minRes <= src.height)
                    .filter(src -> src.height == 0 || src.height == UNKNOWN || maxRes >= src.height)
                    .collect(Collectors.toList());

            if (filteredStreamSources.isEmpty()) {
                throw new ExecutionException(new RuntimeException("No stream left in playlist"));
            } else {
                LOG.debug("{} selected {}", model.getName(), filteredStreamSources.get(filteredStreamSources.size() - 1));
                url = filteredStreamSources.get(filteredStreamSources.size() - 1).getMediaPlaylistUrl();
            }
        }
        LOG.debug("Segment playlist url {}", url);
        return url;
    }

    protected void emptyPlaylistCheck(SegmentPlaylist playlist) {
        if(playlist.segments.isEmpty()) {
            playlistEmptyCount++;
            try {
                Thread.sleep(6000);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        } else {
            playlistEmptyCount = 0;
        }
        if(playlistEmptyCount == 10) {
            LOG.info("Last 10 playlists were empty for {}. Stopping recording!", getModel());
            internalStop();
        }
    }

    abstract void internalStop();

    @Override
    public Model getModel() {
        return model;
    }

    /**
     * Causes the current thread to sleep for a short amount of time.
     * This is used to slow down retries, if something is wrong with the playlist.
     * E.g. HTTP 403 or 404
     */
    protected void waitSomeTime(long waitForMillis) {
        try {
            Thread.sleep(waitForMillis);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            if(running) {
                LOG.error("Couldn't sleep. This might mess up the download!");
            }
        }
    }

    public static class SegmentPlaylist {
        public String url;
        public int seq = 0;
        public float totalDuration = 0;
        public float lastSegDuration = 0;
        public float targetDuration = 0;
        public List<String> segments = new ArrayList<>();
        public boolean encrypted = false;
        public String encryptionMethod = "AES-128";
        public String encryptionKeyUrl;

        public SegmentPlaylist(String url) {
            this.url = url;
        }
    }
}
